@@ -42,6 +42,7 @@ import org.jruby.RubyModule;
 import org.jruby.RubyNumeric;
 import org.jruby.RubyObject;
 import org.jruby.RubyString;
+import org.jruby.anno.JRubyClass;
 import org.jruby.anno.JRubyMethod;
 import org.jruby.exceptions.RaiseException;
 import org.jruby.runtime.Arity;
@@ -64,6 +65,7 @@ import org.jruby.runtime.Visibility;
  *
  * @author <a href="mailto:bill.dortch@gmail.com">Bill Dortch</a>
  */
+@JRubyClass(name = "OpenSSL::BN", include = "Comparable")
 public class BN extends RubyObject {
     private static final long serialVersionUID = -5660938062191525498L;
 
@@ -278,7 +280,7 @@ public class BN extends RubyObject {
     }
 
     @JRubyMethod(name = "to_bn")
-    public IRubyObject to_bn() {
+    public BN to_bn() {
         return this;
     }
 
@@ -850,4 +852,14 @@ public class BN extends RubyObject {
         throw arg.getRuntime().newTypeError("Cannot convert into OpenSSL::BN");
     }
 
+    @Override
+    public Object toJava(Class target) {
+        if ( target.isAssignableFrom(BigInteger.class) || target == Number.class ) return value;
+        if ( target == Long.class || target == Long.TYPE ) return value.longValue();
+        if ( target == Integer.class || target == Integer.TYPE ) return value.intValue();
+        if ( target == Double.class || target == Double.TYPE ) return value.doubleValue();
+        if ( target == Float.class || target == Float.TYPE ) return value.floatValue();
+        return super.toJava(target);
+    }
+
 }
