@@ -41,9 +41,11 @@ import java.security.NoSuchAlgorithmException;
 import java.security.PrivateKey;
 import java.security.cert.CertificateException;
 import java.security.cert.X509Certificate;
+import java.util.Collections;
 
 import javax.net.ssl.KeyManager;
 import javax.net.ssl.SSLEngine;
+import javax.net.ssl.SSLSessionContext;
 import javax.net.ssl.TrustManager;
 import javax.net.ssl.X509ExtendedKeyManager;
 import javax.net.ssl.X509TrustManager;
@@ -83,6 +85,7 @@ import static org.jruby.ext.openssl.X509._X509;
 import static org.jruby.ext.openssl.X509Cert._Certificate;
 import static org.jruby.ext.openssl.OpenSSL.debug;
 import static org.jruby.ext.openssl.OpenSSL.debugStackTrace;
+import static org.jruby.ext.openssl.OpenSSL.warn;
 import static org.jruby.ext.openssl.Utils.hasNonNilInstanceVariable;
 
 /**
@@ -119,15 +122,22 @@ public class SSLContext extends RubyObject {
         SSL_VERSION_OSSL2JSSE.put("SSLv23", "SSL");
         SSL_VERSION_OSSL2JSSE.put("SSLv23_server", "SSL");
         SSL_VERSION_OSSL2JSSE.put("SSLv23_client", "SSL");
-        ENABLED_PROTOCOLS.put("SSL", new String[] { "SSLv2", "SSLv3", "TLSv1" });
+
+        if ( OpenSSL.javaVersion7(true) ) { // >= 1.7
+            ENABLED_PROTOCOLS.put("SSL", new String[] { "SSLv2", "SSLv3", "TLSv1", "TLSv1.1", "TLSv1.2" });
+        } else {
+            ENABLED_PROTOCOLS.put("SSL", new String[] { "SSLv2", "SSLv3", "TLSv1" });
+        }
 
         // Historically we were ahead of MRI to support TLS
         // ... thus the non-standard names version names :
 
         SSL_VERSION_OSSL2JSSE.put("TLS", "TLS");
-        ENABLED_PROTOCOLS.put("TLS", new String[] { "TLSv1", "TLSv1.1" });
+        ENABLED_PROTOCOLS.put("TLS", new String[] { "TLSv1", "TLSv1.1", "TLSv1.2" });
 
         SSL_VERSION_OSSL2JSSE.put("TLSv1.1", "TLSv1.1");
+        SSL_VERSION_OSSL2JSSE.put("TLSv1_1_server", "TLSv1.1");
+        SSL_VERSION_OSSL2JSSE.put("TLSv1_1_client", "TLSv1.1");
         ENABLED_PROTOCOLS.put("TLSv1.1", new String[] { "TLSv1.1" });
 
         SSL_VERSION_OSSL2JSSE.put("TLSv1_1", "TLSv1.1"); // supported on MRI 2.x
@@ -135,6 +145,8 @@ public class SSLContext extends RubyObject {
         ENABLED_PROTOCOLS.put("TLSv1.2", new String[] { "TLSv1.2" });
 
         SSL_VERSION_OSSL2JSSE.put("TLSv1.2", "TLSv1.2"); // just for completeness
+        SSL_VERSION_OSSL2JSSE.put("TLSv1_2_server", "TLSv1.2");
+        SSL_VERSION_OSSL2JSSE.put("TLSv1_2_client", "TLSv1.2");
     }
 
     private static ObjectAllocator SSLCONTEXT_ALLOCATOR = new ObjectAllocator() {
@@ -183,6 +195,9 @@ public class SSLContext extends RubyObject {
         SSLContext.defineConstant("METHODS", methods);
         // in 1.8.7 as well as 1.9.3 :
         // [:TLSv1, :TLSv1_server, :TLSv1_client, :SSLv3, :SSLv3_server, :SSLv3_client, :SSLv23, :SSLv23_server, :SSLv23_client]
+        // in 2.0.0 :
+        // [:TLSv1, :TLSv1_server, :TLSv1_client, :TLSv1_2, :TLSv1_2_server, :TLSv1_2_client, :TLSv1_1, :TLSv1_1_server,
+        //  :TLSv1_1_client, :SSLv3, :SSLv3_server, :SSLv3_client, :SSLv23, :SSLv23_server, :SSLv23_client]
 
         SSLContext.setConstant("SESSION_CACHE_OFF", runtime.newFixnum(SESSION_CACHE_OFF));
         SSLContext.setConstant("SESSION_CACHE_CLIENT", runtime.newFixnum(SESSION_CACHE_CLIENT));
@@ -251,6 +266,10 @@ public class SSLContext extends RubyObject {
         super(runtime,type);
     }
 
+    SSLContext(Ruby runtime) {
+        super(runtime, _SSLContext(runtime));
+    }
+
     private String ciphers = CipherStrings.SSL_DEFAULT_CIPHER_LIST;
     private String protocol = "SSL"; // SSLv23 in OpenSSL by default
     private boolean protocolForServer = true;
@@ -258,29 +277,33 @@ public class SSLContext extends RubyObject {
     private PKey t_key;
     private X509Cert t_cert;
 
-    /* TODO: should move to SSLSession after implemented */
     private int verifyResult = 1; /* avoid 0 (= X509_V_OK) just in case */
 
-    private int sessionCacheMode; // 2
+    //private int sessionCacheMode; // 2 default on MRI
     private int sessionCacheSize; // 20480
 
     private InternalContext internalContext;
 
-    @JRubyMethod(rest = true, visibility = Visibility.PRIVATE)
+    @JRubyMethod(required = 0, optional = 1, visibility = Visibility.PRIVATE)
     public IRubyObject initialize(IRubyObject[] args) {
         if ( args.length > 0 ) set_ssl_version(args[0]);
-        return this;
+        return initializeImpl();
     }
 
+    final SSLContext initializeImpl() { return this; }
+
     @JRubyMethod
     public IRubyObject setup(final ThreadContext context) {
         final Ruby runtime = context.runtime;
 
         if ( isFrozen() ) return runtime.getNil();
 
-        this.freeze(context);
+        synchronized(this) {
+            if ( isFrozen() ) return runtime.getNil();
+            this.freeze(context);
+        }
 
-        internalContext = new InternalContext();
+        final X509Store certStore = getCertStore();
 
         // TODO: handle tmp_dh_callback :
 
@@ -293,16 +316,7 @@ public class SSLContext extends RubyObject {
         //   }
         // #endif
 
-        final X509Store certStore = getCertStore();
-        internalContext.store = certStore != null ? certStore.getStore() : new Store();
-
-        IRubyObject value = getInstanceVariable("@extra_chain_cert");
-        if ( value != null && ! value.isNil() ) {
-            final List<X509Cert> extraCerts = convertToX509Certs(context, value);
-            final ArrayList<X509AuxCertificate> extraChainCert = new ArrayList<X509AuxCertificate>(extraCerts.size());
-            for ( X509Cert x : extraCerts ) extraChainCert.add( x.getAuxCert() );
-            internalContext.extraChainCert = extraChainCert;
-        }
+        IRubyObject value;
 
         value = getInstanceVariable("@key");
         final PKey key;
@@ -326,31 +340,58 @@ public class SSLContext extends RubyObject {
             cert = getCallbackCert(context);
         }
 
-        if ( key != null && cert != null ) {
-            internalContext.keyAlgorithm = key.getAlgorithm();
-            internalContext.privateKey = key.getPrivateKey();
-            internalContext.cert = cert.getAuxCert();
-        }
-
         value = getInstanceVariable("@client_ca");
+        final List<X509AuxCertificate> clientCert;
         if ( value != null && ! value.isNil() ) {
             if ( value.respondsTo("each") ) {
-                for ( X509Cert x : convertToX509Certs(context, value) ) {
-                    internalContext.clientCert.add( x.getAuxCert() );
-                }
+                final List<X509Cert> cCerts = convertToX509Certs(context, value);
+                clientCert = new ArrayList<X509AuxCertificate>(cCerts.size());
+                for ( X509Cert x : cCerts ) clientCert.add( x.getAuxCert() );
             } else {
                 if ( ! ( value instanceof X509Cert ) ) {
                     throw runtime.newTypeError("OpenSSL::X509::Certificate expected but got @client_ca = " + value.inspect());
                 }
-                internalContext.clientCert.add( ((X509Cert) value).getAuxCert() );
+                clientCert = Collections.singletonList( ((X509Cert) value).getAuxCert() );
             }
         }
+        else clientCert = Collections.emptyList();
 
-        String caFile = getCaFile();
-        String caPath = getCaPath();
+        value = getInstanceVariable("@extra_chain_cert");
+        final List<X509AuxCertificate> extraChainCert;
+        if ( value != null && ! value.isNil() ) {
+            final List<X509Cert> eCerts = convertToX509Certs(context, value);
+            extraChainCert = new ArrayList<X509AuxCertificate>(eCerts.size());
+            for ( X509Cert x : eCerts ) extraChainCert.add( x.getAuxCert() );
+        }
+        else {
+            extraChainCert = null;
+        }
+
+        value = getInstanceVariable("@verify_mode");
+        final int verifyMode;
+        if ( value != null && ! value.isNil() ) {
+            verifyMode = RubyNumeric.fix2int(value);
+        }
+        else {
+            verifyMode = SSL.VERIFY_NONE; // 0x00
+        }
+
+        value = getInstanceVariable("@timeout");
+        final int timeout;
+        if ( value != null && ! value.isNil() ) {
+            timeout = RubyNumeric.fix2int(value);
+        }
+        else {
+            timeout = 0;
+        }
+
+        final Store store = certStore != null ? certStore.getStore() : new Store();
+
+        final String caFile = getCaFile();
+        final String caPath = getCaPath();
         if (caFile != null || caPath != null) {
             try {
-                if (internalContext.store.loadLocations(runtime, caFile, caPath) == 0) {
+                if (store.loadLocations(runtime, caFile, caPath) == 0) {
                     runtime.getWarnings().warn(ID.MISCELLANEOUS, "can't set verify locations");
                 }
             }
@@ -360,28 +401,18 @@ public class SSLContext extends RubyObject {
             }
         }
 
-        value = getInstanceVariable("@timeout");
-        if ( value != null && ! value.isNil() ) {
-            internalContext.timeout = RubyNumeric.fix2int(value);
-        }
-
-        value = getInstanceVariable("@verify_mode");
-        if ( value != null && ! value.isNil() ) {
-            internalContext.verifyMode = RubyNumeric.fix2int(value);
-        }
-
         value = getInstanceVariable("@verify_callback");
         if ( value != null && ! value.isNil() ) {
-            internalContext.store.setExtraData(1, value);
+            store.setExtraData(1, value);
         } else {
-            internalContext.store.setExtraData(1, null);
+            store.setExtraData(1, null);
         }
 
         value = getInstanceVariable("@verify_depth");
         if ( value != null && ! value.isNil() ) {
-            internalContext.store.setDepth(RubyNumeric.fix2int(value));
+            store.setDepth(RubyNumeric.fix2int(value));
         } else {
-            internalContext.store.setDepth(-1);
+            store.setDepth(-1);
         }
 
         value = getInstanceVariable("@servername_cb");
@@ -389,7 +420,8 @@ public class SSLContext extends RubyObject {
             // SSL_CTX_set_tlsext_servername_callback(ctx, ssl_servername_cb);
         }
 
-        /* TODO: should be implemented for SSLSession
+        // NOTE: no API under javax.net to support session get/new/remove callbacks
+        /*
         val = ossl_sslctx_get_sess_id_ctx(self);
         if (!NIL_P(val)){
             StringValue(val);
@@ -414,11 +446,14 @@ public class SSLContext extends RubyObject {
         */
 
         try {
-            internalContext.init();
+            internalContext = new InternalContext(
+                cert, key, store, clientCert, extraChainCert, verifyMode, timeout
+            );
         }
         catch (GeneralSecurityException e) {
             throw newSSLError(runtime, e);
         }
+
         return runtime.getTrue();
     }
 
@@ -497,6 +532,8 @@ public class SSLContext extends RubyObject {
         return version;
     }
 
+    final String getProtocol() { return this.protocol; }
+
     // ##
     // # Sets the parameters for this SSL context to the values in +params+.
     // # The keys in +params+ must be assignment methods on SSLContext.
@@ -545,17 +582,18 @@ public class SSLContext extends RubyObject {
         return params;
     }
 
-    // NOTE: mostly stubs since SSL session (cache) not yet implemented :
-
     @JRubyMethod(name = "session_cache_mode")
     public IRubyObject session_cache_mode() {
-        return getRuntime().newFixnum(sessionCacheMode);
+        return getRuntime().getNil();
+        //return getRuntime().newFixnum(sessionCacheMode);
     }
 
     @JRubyMethod(name = "session_cache_mode=")
     public IRubyObject set_session_cache_mode(IRubyObject mode) {
-        this.sessionCacheMode = RubyInteger.fix2int(mode);
-        return mode;
+        //this.sessionCacheMode = RubyInteger.fix2int(mode);
+        //return mode;
+        warn(getRuntime().getCurrentContext(), "SSLContext#session_cache_mode= has no effect under JRuby");
+        return session_cache_mode();
     }
 
     @JRubyMethod(name = "session_cache_size")
@@ -571,7 +609,7 @@ public class SSLContext extends RubyObject {
 
     @JRubyMethod(name = "session_cache_stats")
     public RubyHash session_cache_stats(final ThreadContext context) {
-        // TODO: session cache NOT IMPLEMENTED
+        // NOTE: session cache NOT IMPLEMENTED
 
         // { :connect_renegotiate=>0, :cache_full=>0, :accept_good=>0,
         //   :connect=>0, :timeouts=>0, :accept_renegotiate=>0, :accept=>0,
@@ -787,52 +825,87 @@ public class SSLContext extends RubyObject {
      */
     private class InternalContext {
 
-        Store store;
-        int verifyMode = SSL.VERIFY_NONE; // 0x00
-        X509AuxCertificate cert; String keyAlgorithm; PrivateKey privateKey;
+        InternalContext(
+            final X509Cert xCert,
+            final PKey pKey,
+            final Store store,
+            final List<X509AuxCertificate> clientCert,
+            final List<X509AuxCertificate> extraChainCert,
+            final int verifyMode,
+            final int timeout) throws NoSuchAlgorithmException, KeyManagementException {
 
-        final List<X509AuxCertificate> clientCert = new ArrayList<X509AuxCertificate>();
-        List<X509AuxCertificate> extraChainCert;
-
-        int timeout = 0;
-
-        private javax.net.ssl.SSLContext sslContext;
-
-        void init() throws GeneralSecurityException {
-            this.sslContext = SecurityHelper.getSSLContext(protocol);
-            if (protocolForClient) {
-                sslContext.getClientSessionContext().setSessionTimeout(timeout);
+            if ( pKey != null && xCert != null ) {
+                this.privateKey = pKey.getPrivateKey();
+                this.keyAlgorithm = pKey.getAlgorithm();
+                this.cert = xCert.getAuxCert();
             }
-            if (protocolForServer) {
-                sslContext.getServerSessionContext().setSessionTimeout(timeout);
+            else {
+                this.privateKey = null;
+                this.keyAlgorithm = null;
+                this.cert = null;
             }
-            sslContext.init(
-                new KeyManager[] { new KeyManagerImpl(this) },
-                new TrustManager[] { new TrustManagerImpl(this) },
-                null
-            );
+
+            this.store = store;
+            this.clientCert = clientCert;
+            this.extraChainCert = extraChainCert;
+            this.verifyMode = verifyMode;
+            //this.timeout = timeout;
+
+            // initialize SSL context :
+
+            final javax.net.ssl.SSLContext sslContext;
+            sslContext = SecurityHelper.getSSLContext(protocol);
+
+            if ( protocolForClient ) {
+                final SSLSessionContext clientContext = sslContext.getClientSessionContext();
+                clientContext.setSessionTimeout(timeout);
+                if ( sessionCacheSize >= 0 ) {
+                    clientContext.setSessionCacheSize(sessionCacheSize);
+                }
+            }
+            if ( protocolForServer ) {
+                final SSLSessionContext serverContext = sslContext.getClientSessionContext();
+                serverContext.setSessionTimeout(timeout);
+                if ( sessionCacheSize >= 0 ) {
+                    serverContext.setSessionCacheSize(sessionCacheSize);
+                }
+            }
+
+            final KeyManager[] keyManager = new KeyManager[] { new KeyManagerImpl(this) };
+            final TrustManager[] trustManager = new TrustManager[] { new TrustManagerImpl(this) };
+            sslContext.init(keyManager, trustManager, null);
+            this.sslContext = sslContext;
         }
 
+        final Store store;
+        final X509AuxCertificate cert;
+        final String keyAlgorithm;
+        final PrivateKey privateKey;
+
+        final int verifyMode;
+
+        final List<X509AuxCertificate> clientCert; // assumed always != null
+        final List<X509AuxCertificate> extraChainCert; // empty assumed == null
+
+        //final int timeout;
+
+        private final javax.net.ssl.SSLContext sslContext;
+
         // part of ssl_verify_cert_chain
         StoreContext createStoreContext(final String purpose) {
             if ( store == null ) return null;
 
-            final StoreContext storeContext = new StoreContext();
-            if ( storeContext.init(store, null, null) == 0 ) {
-                return null;
-            }
+            final StoreContext storeContext = new StoreContext(store);
+            if ( storeContext.init(null, null) == 0 ) return null;
+
             // for verify_cb
             storeContext.setExtraData(1, store.getExtraData(1));
-            if ( purpose != null ) {
-                storeContext.setDefault(purpose);
-            }
+            if ( purpose != null ) storeContext.setDefault(purpose);
             storeContext.verifyParameter.inherit(store.verifyParameter);
             return storeContext;
         }
 
-        javax.net.ssl.SSLContext getSSLContext() {
-            return sslContext;
-        }
+        final javax.net.ssl.SSLContext getSSLContext() { return sslContext; }
 
         void setLastVerifyResult(int lastVerifyResult) {
             SSLContext.this.setLastVerifyResult(lastVerifyResult);
