@@ -30,6 +30,8 @@ package org.jruby.ext.openssl;
 import java.io.IOException;
 import java.net.Socket;
 import java.nio.ByteBuffer;
+import java.nio.channels.Channel;
+import java.nio.channels.ClosedChannelException;
 import java.nio.channels.SelectableChannel;
 import java.nio.channels.SelectionKey;
 import java.nio.channels.Selector;
@@ -45,7 +47,6 @@ import javax.net.ssl.SSLEngineResult;
 import javax.net.ssl.SSLException;
 import javax.net.ssl.SSLHandshakeException;
 import javax.net.ssl.SSLPeerUnverifiedException;
-import javax.net.ssl.SSLSession;
 
 import org.jruby.Ruby;
 import org.jruby.RubyArray;
@@ -67,7 +68,6 @@ import org.jruby.runtime.builtin.IRubyObject;
 import org.jruby.util.ByteList;
 import org.jruby.runtime.Visibility;
 
-import static org.jruby.ext.openssl.SSL._SSL;
 import static org.jruby.ext.openssl.SSL.newSSLErrorWaitReadable;
 import static org.jruby.ext.openssl.SSL.newSSLErrorWaitWritable;
 import static org.jruby.ext.openssl.OpenSSL.*;
@@ -85,15 +85,15 @@ public class SSLSocket extends RubyObject {
         }
     };
 
-    public static void createSSLSocket(final Ruby runtime, final RubyModule _SSL) { // OpenSSL::SSL
+    public static void createSSLSocket(final Ruby runtime, final RubyModule SSL) { // OpenSSL::SSL
         final ThreadContext context = runtime.getCurrentContext();
-        RubyClass _SSLSocket = _SSL.defineClassUnder("SSLSocket", runtime.getObject(), SSLSOCKET_ALLOCATOR);
-        _SSLSocket.addReadWriteAttribute(context, "io");
-        _SSLSocket.addReadWriteAttribute(context, "context");
-        _SSLSocket.addReadWriteAttribute(context, "sync_close");
-        _SSLSocket.addReadWriteAttribute(context, "hostname");
-        _SSLSocket.defineAlias("to_io", "io");
-        _SSLSocket.defineAnnotatedMethods(SSLSocket.class);
+        RubyClass SSLSocket = SSL.defineClassUnder("SSLSocket", runtime.getObject(), SSLSOCKET_ALLOCATOR);
+        // SSLSocket.addReadAttribute(context, "io");
+        // SSLSocket.defineAlias("to_io", "io");
+        // SSLSocket.addReadAttribute(context, "context");
+        SSLSocket.addReadWriteAttribute(context, "sync_close");
+        SSLSocket.addReadWriteAttribute(context, "hostname");
+        SSLSocket.defineAnnotatedMethods(SSLSocket.class);
     }
 
     public SSLSocket(Ruby runtime, RubyClass type) {
@@ -137,14 +137,18 @@ public class SSLSocket extends RubyObject {
 
     int verifyResult = X509Utils.V_OK;
 
-    @JRubyMethod(name = "initialize", rest = true, frame = true, visibility = Visibility.PRIVATE)
+    @Deprecated
     public IRubyObject _initialize(final ThreadContext context,
         final IRubyObject[] args, final Block unused) {
+        return initialize(context, args);
+    }
+
+    @JRubyMethod(name = "initialize", rest = true, frame = true, visibility = Visibility.PRIVATE)
+    public IRubyObject initialize(final ThreadContext context, final IRubyObject[] args) {
         final Ruby runtime = context.runtime;
 
         if ( Arity.checkArgumentCount(runtime, args, 1, 2) == 1 ) {
-            RubyModule _SSLContext = _SSL(runtime).getClass("SSLContext");
-            sslContext = (SSLContext) _SSLContext.callMethod(context, "new");
+            sslContext = new SSLContext(runtime).initializeImpl();
         } else {
             sslContext = (SSLContext) args[1];
         }
@@ -152,15 +156,15 @@ public class SSLSocket extends RubyObject {
         if ( ! ( args[0] instanceof RubyIO ) ) {
             throw runtime.newTypeError("IO expected but got " + args[0].getMetaClass().getName());
         }
-        this.callMethod(context, "io=", this.io = (RubyIO) args[0]);
-        this.callMethod(context, "context=", sslContext);
+        setInstanceVariable("@io", this.io = (RubyIO) args[0]); // compat (we do not read @io)
+        setInstanceVariable("@context", this.sslContext); // only compat (we do not use @context)
         // This is a bit of a hack: SSLSocket should share code with
         // RubyBasicSocket, which always sets sync to true.
         // Instead we set it here for now.
-        this.io.callMethod(context, "sync=", runtime.getTrue());
+        this.set_sync(context, runtime.getTrue()); // io.sync = true
         this.callMethod(context, "sync_close=", runtime.getFalse());
         sslContext.setup(context);
-        return Utils.invokeSuper(context, this, args, unused); // super()
+        return Utils.invokeSuper(context, this, args, Block.NULL_BLOCK); // super()
     }
 
     private SSLEngine ossl_ssl_setup(final ThreadContext context)
@@ -168,13 +172,13 @@ public class SSLSocket extends RubyObject {
         SSLEngine engine = this.engine;
         if ( engine != null ) return engine;
 
-        final Socket socket = getSocketChannel().socket();
         // Server Name Indication (SNI) RFC 3546
         // SNI support will not be attempted unless hostname is explicitly set by the caller
         String peerHost = this.callMethod(context, "hostname").toString();
-        int peerPort = socket.getPort();
+        final int peerPort = socketChannelImpl().getRemotePort();
         engine = sslContext.createSSLEngine(peerHost, peerPort);
-        final SSLSession session = engine.getSession();
+
+        final javax.net.ssl.SSLSession session = engine.getSession();
         peerNetData = ByteBuffer.allocate(session.getPacketBufferSize());
         peerAppData = ByteBuffer.allocate(session.getApplicationBufferSize());
         netData = ByteBuffer.allocate(session.getPacketBufferSize());
@@ -185,17 +189,33 @@ public class SSLSocket extends RubyObject {
         return this.engine = engine;
     }
 
+    @JRubyMethod(name = "io", alias = "to_io")
+    public final RubyIO io() { return this.io; }
+
+    @JRubyMethod(name = "context")
+    public final SSLContext context() { return this.sslContext; }
+
+    @JRubyMethod(name = "sync")
+    public IRubyObject sync(final ThreadContext context) {
+        return this.io.callMethod(context, "sync");
+    }
+
+    @JRubyMethod(name = "sync=")
+    public IRubyObject set_sync(final ThreadContext context, final IRubyObject sync) {
+        return this.io.callMethod(context, "sync=", sync);
+    }
+
     @JRubyMethod
-    public IRubyObject connect(ThreadContext context) {
-        return connectCommon(context, true);
+    public IRubyObject connect(final ThreadContext context) {
+        return connectImpl(context, true);
     }
 
     @JRubyMethod
     public IRubyObject connect_nonblock(ThreadContext context) {
-        return connectCommon(context, false);
+        return connectImpl(context, false);
     }
 
-    private IRubyObject connectCommon(final ThreadContext context, boolean blocking) {
+    private SSLSocket connectImpl(final ThreadContext context, final boolean blocking) {
         final Ruby runtime = context.runtime;
 
         if ( ! sslContext.isProtocolForClient() ) {
@@ -213,20 +233,24 @@ public class SSLSocket extends RubyObject {
             doHandshake(blocking);
         }
         catch (SSLHandshakeException e) {
+            //debugStackTrace(runtime, e);
             // unlike server side, client should close outbound channel even if
             // we have remaining data to be sent.
             forceClose();
             throw newSSLErrorFromHandshake(runtime, e);
         }
         catch (NoSuchAlgorithmException e) {
+            debugStackTrace(runtime, e);
             forceClose();
             throw newSSLError(runtime, e);
         }
         catch (KeyManagementException e) {
+            debugStackTrace(runtime, e);
             forceClose();
             throw newSSLError(runtime, e);
         }
         catch (IOException e) {
+            //debugStackTrace(runtime, e);
             forceClose();
             throw newSSLError(runtime, e);
         }
@@ -235,15 +259,20 @@ public class SSLSocket extends RubyObject {
 
     @JRubyMethod
     public IRubyObject accept(ThreadContext context) {
-        return acceptCommon(context, true);
+        return acceptImpl(context, true);
     }
 
     @JRubyMethod
     public IRubyObject accept_nonblock(ThreadContext context) {
-        return acceptCommon(context, false);
+        return acceptImpl(context, false);
     }
 
-    public SSLSocket acceptCommon(final ThreadContext context, boolean blocking) {
+    @Deprecated
+    public SSLSocket acceptCommon(ThreadContext context, boolean blocking) {
+        return acceptImpl(context, blocking);
+    }
+
+    private SSLSocket acceptImpl(final ThreadContext context, final boolean blocking) {
         final Ruby runtime = context.runtime;
 
         if ( ! sslContext.isProtocolForServer() ) {
@@ -252,20 +281,19 @@ public class SSLSocket extends RubyObject {
 
         try {
             if ( ! initialHandshake ) {
-                SSLEngine engine = ossl_ssl_setup(context);
+                final SSLEngine engine = ossl_ssl_setup(context);
                 engine.setUseClientMode(false);
-                final IRubyObject verify_mode;
-                if( ! sslContext.isNil() &&
-                    ! ( verify_mode = sslContext.callMethod(context, "verify_mode") ).isNil() ) {
-                    final int vfy = RubyNumeric.fix2int(verify_mode);
-                    if ( vfy == 0 ) { // VERIFY_NONE
+                final IRubyObject verify_mode = sslContext.callMethod(context, "verify_mode");
+                if ( ! verify_mode.isNil() ) {
+                    final int verify = RubyNumeric.fix2int(verify_mode);
+                    if ( verify == 0 ) { // VERIFY_NONE
                         engine.setNeedClientAuth(false);
                         engine.setWantClientAuth(false);
                     }
-                    if ( ( vfy & 1 ) != 0 ) { // VERIFY_PEER
+                    if ( ( verify & 1 ) != 0 ) { // VERIFY_PEER
                         engine.setWantClientAuth(true);
                     }
-                    if ( ( vfy & 2 ) != 0 ) { // VERIFY_FAIL_IF_NO_PEER_CERT
+                    if ( ( verify & 2 ) != 0 ) { // VERIFY_FAIL_IF_NO_PEER_CERT
                         engine.setNeedClientAuth(true);
                     }
                 }
@@ -276,15 +304,38 @@ public class SSLSocket extends RubyObject {
             doHandshake(blocking);
         }
         catch (SSLHandshakeException e) {
+            final String msg = e.getMessage();
+            // updated JDK (>= 1.7.0_75) with deprecated SSL protocols :
+            // javax.net.ssl.SSLHandshakeException: No appropriate protocol (protocol is disabled or cipher suites are inappropriate)
+            if ( e.getCause() == null && msg != null &&
+                 msg.contains("(protocol is disabled or cipher suites are inappropriate)") )  {
+                debug(runtime, sslContext.getProtocol() + " protocol has been deactivated and is not available by default\n see the java.security.Security property jdk.tls.disabledAlgorithms in <JRE_HOME>/lib/security/java.security file");
+            }
+            else {
+                debugStackTrace(runtime, e);
+            }
             throw newSSLErrorFromHandshake(runtime, e);
         }
         catch (NoSuchAlgorithmException e) {
+            debugStackTrace(runtime, e);
             throw newSSLError(runtime, e);
         }
         catch (KeyManagementException e) {
+            debugStackTrace(runtime, e);
             throw newSSLError(runtime, e);
         }
         catch (IOException e) {
+            debugStackTrace(runtime, e);
+            throw newSSLError(runtime, e);
+        }
+        catch (RaiseException e) {
+            throw e;
+        }
+        catch (RuntimeException e) {
+            debugStackTrace(runtime, e);
+            if ( "Could not generate DH keypair".equals( e.getMessage() ) ) {
+                throw SSL.handleCouldNotGenerateDHKeyPairError(runtime, e);
+            }
             throw newSSLError(runtime, e);
         }
         return this;
@@ -306,16 +357,15 @@ public class SSLSocket extends RubyObject {
     // temporarily. SSLSocket requires wrapping IO to be selectable so it should
     // be OK to set configureBlocking(false) permanently.
     private boolean waitSelect(final int operations, final boolean blocking) throws IOException {
-        if ( ! ( io.getChannel() instanceof SelectableChannel ) ) {
-            return true;
-        }
+        final SocketChannelImpl channel = socketChannelImpl();
+        if ( ! channel.isSelectable() ) return true;
+
         final Ruby runtime = getRuntime();
         final RubyThread thread = runtime.getCurrentContext().getThread();
 
-        SelectableChannel selectable = (SelectableChannel)io.getChannel();
-        selectable.configureBlocking(false);
+        channel.configureBlocking(false);
         final Selector selector = runtime.getSelectorPool().get();
-        final SelectionKey key = selectable.register(selector, operations);
+        final SelectionKey key = channel.register(selector, operations);
 
         try {
             io.addBlockingThread(thread);
@@ -410,7 +460,7 @@ public class SSLSocket extends RubyObject {
         throw newSSLErrorWaitWritable(runtime, "write would block");
     }
 
-    private void doHandshake(boolean blocking) throws IOException {
+    private void doHandshake(final boolean blocking) throws IOException {
         while (true) {
             boolean ready = waitSelect(SelectionKey.OP_READ | SelectionKey.OP_WRITE, blocking);
 
@@ -482,7 +532,7 @@ public class SSLSocket extends RubyObject {
     private int writeToChannel(ByteBuffer buffer, boolean blocking) throws IOException {
         int totalWritten = 0;
         while ( buffer.hasRemaining() ) {
-            totalWritten += getSocketChannel().write(buffer);
+            totalWritten += socketChannelImpl().write(buffer);
             if ( ! blocking ) break; // don't continue attempting to read
         }
         return totalWritten;
@@ -497,29 +547,29 @@ public class SSLSocket extends RubyObject {
             throw new IOException("Writing not possible during handshake");
         }
 
-        SelectableChannel selectable = getSocketChannel();
-        boolean blockingMode = selectable.isBlocking();
-        if ( ! blocking ) selectable.configureBlocking(false);
+        SocketChannelImpl channel = socketChannelImpl();
+        final boolean blockingMode = channel.isBlocking();
+        if ( ! blocking ) channel.configureBlocking(false);
 
         try {
             if ( netData.hasRemaining() ) {
                 flushData(blocking);
             }
             netData.clear();
-            SSLEngineResult res = engine.wrap(src, netData);
-            if (res.getStatus()==SSLEngineResult.Status.CLOSED) {
-            	  throw getRuntime().newIOError("closed SSL engine");
+            final SSLEngineResult result = engine.wrap(src, netData);
+            if ( result.getStatus() == SSLEngineResult.Status.CLOSED ) {
+                throw getRuntime().newIOError("closed SSL engine");
             }
             netData.flip();
             flushData(blocking);
-            return res.bytesConsumed();
+            return result.bytesConsumed();
         }
         finally {
-            if ( ! blocking ) selectable.configureBlocking(blockingMode);
+            if ( ! blocking ) channel.configureBlocking(blockingMode);
         }
     }
 
-    public int read(ByteBuffer dst, boolean blocking) throws IOException {
+    public int read(final ByteBuffer dst, final boolean blocking) throws IOException {
         if ( initialHandshake ) return 0;
         if ( engine.isInboundDone() ) return -1;
 
@@ -535,10 +585,11 @@ public class SSLSocket extends RubyObject {
         return limit;
     }
 
-    private int readAndUnwrap(boolean blocking) throws IOException {
-        int bytesRead = getSocketChannel().read(peerNetData);
-        if (bytesRead == -1) {
-            if (!peerNetData.hasRemaining() || (status == SSLEngineResult.Status.BUFFER_UNDERFLOW)) {
+    private int readAndUnwrap(final boolean blocking) throws IOException {
+        final int bytesRead = socketChannelImpl().read(peerNetData);
+        if ( bytesRead == -1 ) {
+            if ( ! peerNetData.hasRemaining() ||
+                 ( status == SSLEngineResult.Status.BUFFER_UNDERFLOW ) ) {
                 closeInbound();
                 return -1;
             }
@@ -548,23 +599,27 @@ public class SSLSocket extends RubyObject {
         }
         peerAppData.clear();
         peerNetData.flip();
-        SSLEngineResult res;
+
+        SSLEngineResult result;
         do {
-            res = engine.unwrap(peerNetData, peerAppData);
-        } while (res.getStatus() == SSLEngineResult.Status.OK &&
-				res.getHandshakeStatus() == SSLEngineResult.HandshakeStatus.NEED_UNWRAP &&
-				res.bytesProduced() == 0);
-        if(res.getHandshakeStatus() == SSLEngineResult.HandshakeStatus.FINISHED) {
+            result = engine.unwrap(peerNetData, peerAppData);
+        }
+        while ( result.getStatus() == SSLEngineResult.Status.OK &&
+				result.getHandshakeStatus() == SSLEngineResult.HandshakeStatus.NEED_UNWRAP &&
+				result.bytesProduced() == 0 );
+
+        if ( result.getHandshakeStatus() == SSLEngineResult.HandshakeStatus.FINISHED ) {
             finishInitialHandshake();
         }
-        if(peerAppData.position() == 0 &&
-            res.getStatus() == SSLEngineResult.Status.OK &&
-            peerNetData.hasRemaining()) {
-            res = engine.unwrap(peerNetData, peerAppData);
+        if ( peerAppData.position() == 0 &&
+            result.getStatus() == SSLEngineResult.Status.OK &&
+            peerNetData.hasRemaining() ) {
+            result = engine.unwrap(peerNetData, peerAppData);
         }
-        status = res.getStatus();
-        handshakeStatus = res.getHandshakeStatus();
-        if ( bytesRead == -1 && !peerNetData.hasRemaining() ) {
+        status = result.getStatus();
+        handshakeStatus = result.getHandshakeStatus();
+
+        if ( bytesRead == -1 && ! peerNetData.hasRemaining() ) {
             // now it's safe to call closeInbound().
             closeInbound();
         }
@@ -572,11 +627,13 @@ public class SSLSocket extends RubyObject {
             doShutdown();
             return -1;
         }
+
         peerNetData.compact();
         peerAppData.flip();
-        if(!initialHandshake && (handshakeStatus == SSLEngineResult.HandshakeStatus.NEED_TASK ||
-                                 handshakeStatus == SSLEngineResult.HandshakeStatus.NEED_WRAP ||
-                                 handshakeStatus == SSLEngineResult.HandshakeStatus.FINISHED)) {
+        if ( ! initialHandshake && (
+                handshakeStatus == SSLEngineResult.HandshakeStatus.NEED_TASK ||
+                handshakeStatus == SSLEngineResult.HandshakeStatus.NEED_WRAP ||
+                handshakeStatus == SSLEngineResult.HandshakeStatus.FINISHED ) ) {
             doHandshake(blocking);
         }
         return peerAppData.remaining();
@@ -612,7 +669,7 @@ public class SSLSocket extends RubyObject {
         flushData(true);
     }
 
-    private IRubyObject do_sysread(final ThreadContext context,
+    private RubyString sysreadImpl(final ThreadContext context,
         final IRubyObject[] args, final boolean blocking) {
         final Ruby runtime = context.runtime;
 
@@ -643,7 +700,7 @@ public class SSLSocket extends RubyObject {
             // ensure >0 bytes read; sysread is blocking read.
             while ( rr <= 0 ) {
                 if ( engine == null ) {
-                    rr = getSocketChannel().read(dst);
+                    rr = socketChannelImpl().read(dst);
                 } else {
                     rr = read(dst, blocking);
                 }
@@ -671,13 +728,13 @@ public class SSLSocket extends RubyObject {
 
     @JRubyMethod(rest = true, required = 1, optional = 1)
     public IRubyObject sysread(ThreadContext context, IRubyObject[] args) {
-        return do_sysread(context, args, true);
+        return sysreadImpl(context, args, true);
     }
 
     @JRubyMethod(rest = true, required = 1, optional = 2)
     public IRubyObject sysread_nonblock(ThreadContext context, IRubyObject[] args) {
         // TODO: options for exception raising
-        return do_sysread(context, args, false);
+        return sysreadImpl(context, args, false);
     }
 
     private IRubyObject do_syswrite(final ThreadContext context,
@@ -697,7 +754,7 @@ public class SSLSocket extends RubyObject {
                 written = write(b1, blocking);
             }
 
-            this.callMethod(context, "io").callMethod(context, "flush");
+            this.io.callMethod(context, "flush");
 
             return runtime.newFixnum(written);
         }
@@ -723,7 +780,7 @@ public class SSLSocket extends RubyObject {
     }
 
     private void checkClosed() {
-        if ( ! getSocketChannel().isOpen() ) {
+        if ( ! socketChannelImpl().isOpen() ) {
             throw getRuntime().newIOError("closed stream");
         }
     }
@@ -756,7 +813,7 @@ public class SSLSocket extends RubyObject {
         close( sslContext.isProtocolForClient() );
 
         if ( this.callMethod(context, "sync_close").isTrue() ) {
-            this.callMethod(context, "io").callMethod(context, "close");
+            this.io.callMethod(context, "close");
         }
         return context.runtime.getNil();
     }
@@ -848,13 +905,120 @@ public class SSLSocket extends RubyObject {
         return getRuntime().getNil(); // throw new UnsupportedOperationException();
     }
 
+    javax.net.ssl.SSLSession getSession() {
+        return engine == null ? null : engine.getSession();
+    }
+
+    private transient SSLSession session;
+
+    @JRubyMethod(name = "session")
+    public IRubyObject session(final ThreadContext context) {
+        if ( getSession() == null ) return context.nil;
+        if ( session == null ) {
+            return session = new SSLSession(context.runtime).initializeImpl(context, this);
+        }
+        return session;
+    }
+
     @JRubyMethod(name = "session=")
     public IRubyObject set_session(IRubyObject session) {
         warn(getRuntime().getCurrentContext(), "WARNING: SSLSocket#session= is not supported");
         return getRuntime().getNil(); // throw new UnsupportedOperationException();
     }
 
-    private SocketChannel getSocketChannel() {
-        return (SocketChannel) io.getChannel();
+    @JRubyMethod
+    public IRubyObject ssl_version() {
+        if ( engine == null ) return getRuntime().getNil();
+        return getRuntime().newString( engine.getSession().getProtocol() );
     }
+
+    private transient SocketChannelImpl socketChannel;
+
+    private SocketChannelImpl socketChannelImpl() {
+        if ( socketChannel != null ) return socketChannel;
+
+        final Channel channel = io.getChannel();
+        if ( channel instanceof SocketChannel ) {
+            return socketChannel = new JavaSocketChannel((SocketChannel) channel);
+        }
+
+        // TODO JNR
+
+        throw new IllegalStateException("unknow channel impl: " + channel + " of type " + channel.getClass().getName());
+    }
+
+    private static interface SocketChannelImpl {
+
+        boolean isOpen() ;
+
+        int read(ByteBuffer dst) throws IOException ;
+
+        int write(ByteBuffer src) throws IOException ;
+
+        int getRemotePort();
+
+        boolean isSelectable() ;
+
+        // SelectableChannel
+
+        boolean isBlocking() ;
+
+        void configureBlocking(boolean block) throws IOException ;
+
+        SelectionKey register(Selector selector, int ops) throws IOException ;
+
+        //boolean selectionOpsReadable(final int readyOps);
+
+        //boolean selectionOpsWritable(final int readyOps) ;
+
+    }
+
+    private static final class JavaSocketChannel implements SocketChannelImpl {
+
+        JavaSocketChannel(final SocketChannel channel) {
+            this.channel = channel;
+        }
+
+        private final SocketChannel channel;
+
+        public boolean isOpen() { return channel.isOpen(); }
+
+        public int read(ByteBuffer dst) throws IOException {
+            return channel.read(dst);
+        }
+
+        public int write(ByteBuffer src) throws IOException {
+            return channel.write(src);
+        }
+
+        public int getRemotePort() { return channel.socket().getPort(); }
+
+        public boolean isSelectable() {
+            return true; // return channel instanceof SelectableChannel;
+        }
+
+        public boolean isBlocking() { return channel.isBlocking(); }
+
+        public void configureBlocking(boolean block) throws IOException {
+            channel.configureBlocking(block);
+        }
+
+        public SelectionKey register(Selector selector, int ops) throws ClosedChannelException {
+            return channel.register(selector, ops);
+        }
+
+        public boolean selectionOpsReadable(final int readyOps) {
+            return (readyOps & SelectionKey.OP_READ) != 0;
+        }
+
+        public boolean selectionOpsWritable(final int readyOps) {
+            return (readyOps & SelectionKey.OP_WRITE) != 0;
+        }
+
+    }
+
+    private static boolean jnrChannel(final Channel channel) {
+        return channel.getClass().getName().startsWith("jnr.");
+    }
+
 }// SSLSocket
