@@ -27,6 +27,7 @@
  ***** END LICENSE BLOCK *****/
 package org.jruby.ext.openssl.x509store;
 
+import org.jruby.ext.openssl.util.Cache;
 import static org.jruby.ext.openssl.x509store.X509Utils.X509_CERT_DIR;
 import static org.jruby.ext.openssl.x509store.X509Utils.X509_FILETYPE_ASN1;
 import static org.jruby.ext.openssl.x509store.X509Utils.X509_FILETYPE_DEFAULT;
@@ -56,6 +57,7 @@ import java.math.BigInteger;
 import java.security.GeneralSecurityException;
 import java.security.KeyStore;
 import java.security.cert.CRL;
+import java.security.cert.CertificateException;
 import java.security.cert.PKIXParameters;
 import java.security.cert.TrustAnchor;
 import java.security.cert.X509Certificate;
@@ -67,6 +69,7 @@ import org.jruby.Ruby;
 import org.jruby.RubyHash;
 import org.jruby.ext.openssl.SecurityHelper;
 import org.jruby.util.JRubyFile;
+import org.jruby.util.SafePropertyAccessor;
 import org.jruby.util.io.ChannelDescriptor;
 import org.jruby.util.io.ChannelStream;
 import org.jruby.util.io.FileExistsException;
@@ -153,42 +156,91 @@ public class Lookup {
         return 1;
     }
 
+    private static final Cache<String, Object[]> certCache;
+
+    static {
+        // jruby.openssl.x509.cert.cache = true / false / 8
+        final String cache = SafePropertyAccessor.getProperty("jruby.openssl.x509.lookup.cache");
+        Cache<String, Object[]> certCacheInstance = null;
+        if ( cache != null ) {
+            try {
+                certCacheInstance = Cache.newStrongSoftCache( Integer.parseInt(cache) );
+            }
+            catch (NumberFormatException ex) {
+                if ( Boolean.parseBoolean(cache) ) {
+                    certCacheInstance = Cache.newSoftCache();
+                }
+            }
+        }
+        if ( certCacheInstance == null ) certCacheInstance = Cache.getNullCache();
+        certCache = certCacheInstance;
+    }
+
     /**
      * c: X509_LOOKUP_load_cert_file
      */
-    public int loadCertificateFile(String file, int type) throws Exception {
+    public int loadCertificateFile(final String file, final int type) throws IOException, CertificateException {
         if ( file == null ) return 1;
 
-        int count = 0;
-        int ret = 0;
-        Reader reader = null;
+        final Object[] cached = certCache.get(file);
+
+        BufferedReader reader = null;
         try {
-            InputStream in = wrapJRubyNormalizedInputStream(file);
             X509AuxCertificate auxCert;
-            if (type == X509_FILETYPE_PEM) {
-                reader = new BufferedReader(new InputStreamReader(in));
-                for (;;) {
-                    auxCert = PEMInputOutput.readX509Aux(reader, null);
-                    if ( auxCert == null ) break;
-                    final int i = store.addCertificate(auxCert);
-                    if ( i == 0 ) return ret;
-                    count++;
+            if ( type == X509_FILETYPE_PEM ) {
+                int count = 0;
+                if ( cached != null ) {
+                    boolean storeError = false; for ( int c = 0; c < cached.length; c++ ) {
+                        auxCert = buildAuxFromCached((X509Certificate) cached[c]);
+
+                        if ( ! storeError ) {
+                            if ( store.addCertificate(auxCert) != 0 ) count++;
+                            else { storeError = true; count = 0; } // return 0
+                        }
+                    }
                 }
-                ret = count;
+                else {
+                    reader = new BufferedReader(new InputStreamReader(wrapJRubyNormalizedInputStream(file)));
+                    final ArrayList<Object> cacheEntry = new ArrayList<Object>(8);
+
+                    boolean storeError = false; for (;;) {
+                        auxCert = PEMInputOutput.readX509Aux(reader, null);
+                        if ( auxCert == null ) break;
+
+                        cacheEntry.add( auxCert.cloneForCache() ); // make sure we cache aux
+
+                        if ( ! storeError ) {
+                            if ( store.addCertificate(auxCert) != 0 ) count++;
+                            else { storeError = true; count = 0; } // return 0
+                        }
+                    }
+
+                    certCache.put(file, cacheEntry.toArray( new Object[ cacheEntry.size() ] ));
+                }
+                return count;
             }
-            else if (type == X509_FILETYPE_ASN1) {
-                X509Certificate cert = (X509Certificate)
-                    SecurityHelper.getCertificateFactory("X.509").generateCertificate(in);
-                auxCert = StoreContext.ensureAux(cert);
-                if ( auxCert == null ) {
-                    X509Error.addError(13);
-                    return ret;
+            else if ( type == X509_FILETYPE_ASN1 ) {
+                final X509Certificate cert;
+                if ( cached != null ) {
+                    cert = (X509Certificate) cached[0];
+                    auxCert = buildAuxFromCached(cert);
                 }
-                final int i = store.addCertificate(auxCert);
-                if ( i == 0 ) return ret;
-                ret = i;
-            } else {
+                else {
+                    InputStream in = wrapJRubyNormalizedInputStream(file);
+                    cert = (X509Certificate)
+                        SecurityHelper.getCertificateFactory("X.509").generateCertificate(in);
+
+                    auxCert = new X509AuxCertificate(cert);
+                    certCache.put(file, new Object[] { auxCert.cloneForCache() });
+                }
+                //if ( auxCert == null ) {
+                //    X509Error.addError(13); return 0;
+                //}
+                return store.addCertificate(auxCert);
+            }
+            else {
                 X509Error.addError(X509_R_BAD_X509_FILETYPE);
+                return 0; // NOTE: really?
             }
         }
         finally {
@@ -196,44 +248,44 @@ public class Lookup {
                 try { reader.close(); } catch (Exception ignored) {}
             }
         }
-        return ret;
+    }
+
+    private static X509AuxCertificate buildAuxFromCached(final X509Certificate cached) {
+        X509AuxCertificate auxCert = StoreContext.ensureAux(cached);
+        if ( cached == auxCert ) auxCert = auxCert.clone();
+        return auxCert;
     }
 
     /**
      * c: X509_LOOKUP_load_crl_file
      */
-    public int loadCRLFile(String file, int type) throws Exception {
+    public int loadCRLFile(final String file, final int type) throws Exception {
         if ( file == null ) return 1;
 
-        int count = 0;
-        int ret = 0;
-        Reader reader = null;
+        BufferedReader reader = null;
         try {
             InputStream in = wrapJRubyNormalizedInputStream(file);
             CRL crl;
-            if (type == X509_FILETYPE_PEM) {
+            if ( type == X509_FILETYPE_PEM ) {
                 reader = new BufferedReader(new InputStreamReader(in));
-                for (;;) {
+                int count = 0; for (;;) {
                     crl = PEMInputOutput.readX509CRL(reader, null);
                     if ( crl == null ) break;
-                    final int i = store.addCRL(crl);
-                    if ( i == 0 ) return ret;
+                    if ( store.addCRL(crl) == 0 ) return 0;
                     count++;
                 }
-                ret = count;
+                return count;
             }
-            else if (type == X509_FILETYPE_ASN1) {
+            else if ( type == X509_FILETYPE_ASN1 ) {
                 crl = SecurityHelper.getCertificateFactory("X.509").generateCRL(in);
-                if (crl == null) {
-                    X509Error.addError(13);
-                    return ret;
-                }
-                final int i = store.addCRL(crl);
-                if ( i == 0 ) return ret;
-                ret = i;
+                //if ( crl == null ) {
+                //    X509Error.addError(13); return 0;
+                //}
+                return store.addCRL(crl);
             }
             else {
                 X509Error.addError(X509_R_BAD_X509_FILETYPE);
+                return 0; // NOTE: really?
             }
         }
         finally {
@@ -241,40 +293,63 @@ public class Lookup {
                 try { reader.close(); } catch (Exception ignored) {}
             }
         }
-        return ret;
     }
 
     /**
      * c: X509_LOOKUP_load_cert_crl_file
      */
-    public int loadCertificateOrCRLFile(String file, int type) throws Exception {
+    public int loadCertificateOrCRLFile(final String file, final int type) throws IOException, CertificateException {
         if ( type != X509_FILETYPE_PEM ) return loadCertificateFile(file, type);
 
-        int count = 0;
-        Reader reader = null;
-        try {
-            InputStream in = wrapJRubyNormalizedInputStream(file);
-            reader = new BufferedReader(new InputStreamReader(in));
-            for (;;) {
-                Object v = PEMInputOutput.readPEM(reader, null);
-                if ( v == null ) break;
+        final Object[] cached = certCache.get(file);
 
-                if ( v instanceof X509Certificate ) {
-                    store.addCertificate(StoreContext.ensureAux((X509Certificate) v));
-                    count++;
-                }
-                else if ( v instanceof CRL ) {
-                    store.addCRL((CRL) v);
-                    count++;
+        BufferedReader reader = null;
+        try {
+            int count = 0;
+            if ( cached != null ) {
+                for ( int c = 0; c < cached.length; c++ ) {
+                    Object cert = cached[c];
+                    if ( cert instanceof X509Certificate ) {
+                        store.addCertificate(buildAuxFromCached((X509Certificate) cert));
+                        count++;
+                    }
+                    else if ( cert instanceof CRL ) {
+                        store.addCRL((CRL) cert);
+                        count++;
+                    }
                 }
             }
+            else {
+                reader = new BufferedReader(new InputStreamReader(wrapJRubyNormalizedInputStream(file)));
+                final ArrayList<Object> cacheEntry = new ArrayList<Object>(8);
+                for (;;) {
+                    Object cert = PEMInputOutput.readPEM(reader, null);
+                    if ( cert == null ) break;
+
+                    if ( cert instanceof X509Certificate ) {
+                        X509AuxCertificate auxCert = StoreContext.ensureAux((X509Certificate) cert);
+                        store.addCertificate(auxCert);
+                        count++;
+
+                        cert = auxCert.cloneForCache(); // make sure we cache aux
+                    }
+                    else if ( cert instanceof CRL ) {
+                        store.addCRL((CRL) cert);
+                        count++;
+                    }
+
+                    cacheEntry.add(cert);
+                }
+
+                certCache.put(file, cacheEntry.toArray( new Object[ cacheEntry.size() ] ));
+            }
+            return count;
         }
         finally {
             if ( reader != null ) {
                 try { reader.close(); } catch (Exception ignored) {}
             }
         }
-        return count;
     }
 
     public int loadDefaultJavaCACertsFile() throws IOException, GeneralSecurityException {
@@ -306,20 +381,23 @@ public class Lookup {
             try {
                 ChannelDescriptor descriptor = ChannelDescriptor.open(runtime.getCurrentDirectory(), file, new ModeFlags(ModeFlags.RDONLY));
                 return ChannelStream.open(runtime, descriptor).newInputStream();
-            } catch (NoSuchMethodError nsme) {
+            }
+            catch (NoSuchMethodError ex) {
                 File f = new File(file);
                 if ( ! f.isAbsolute() ) {
                     f = new File(runtime.getCurrentDirectory(), file);
                 }
                 return new BufferedInputStream(new FileInputStream(f));
-            } catch (FileExistsException fee) {
+            }
+            catch (FileExistsException ex) {
                 // should not happen because ModeFlag does not contain CREAT.
-                fee.printStackTrace(System.err);
-                throw new IllegalStateException(fee.getMessage(), fee);
-            } catch (InvalidValueException ive) {
+                ex.printStackTrace(System.err);
+                throw new IllegalStateException(ex.getMessage(), ex);
+            }
+            catch (InvalidValueException ex) {
                 // should not happen because ModeFlasg does not contain APPEND.
-                ive.printStackTrace(System.err);
-                throw new IllegalStateException(ive.getMessage(), ive);
+                ex.printStackTrace(System.err);
+                throw new IllegalStateException(ex.getMessage(), ex);
             }
         }
     }
